
/*
 * This file was automatically generated by sdbus-c++-xml2cpp; DO NOT EDIT!
 */

#ifndef __sdbuscpp__celsius_thermometer_proxy_h__proxy__H__
#define __sdbuscpp__celsius_thermometer_proxy_h__proxy__H__

#include <sdbus-c++/sdbus-c++.h>
#include <string>
#include <tuple>

namespace org {
namespace sdbuscpp {
namespace stresstests {
namespace celsius {

class thermometer_proxy
{
public:
    static constexpr const char* INTERFACE_NAME = "org.sdbuscpp.stresstests.celsius.thermometer";

protected:
    thermometer_proxy(sdbus::IProxy& proxy)
        : m_proxy(proxy)
    {
    }

    thermometer_proxy(const thermometer_proxy&) = delete;
    thermometer_proxy& operator=(const thermometer_proxy&) = delete;
    thermometer_proxy(thermometer_proxy&&) = delete;
    thermometer_proxy& operator=(thermometer_proxy&&) = delete;

    ~thermometer_proxy() = default;

    void registerProxy()
    {
    }

public:
    uint32_t getCurrentTemperature()
    {
        uint32_t result;
        m_proxy.callMethod("getCurrentTemperature").onInterface(INTERFACE_NAME).storeResultsTo(result);
        return result;
    }

private:
    sdbus::IProxy& m_proxy;
};

}}}} // namespaces

#endif
